#include "dynapro.h"

int    *iVector(long, long);
int   **iMatrix(long, long, long, long);
void free_iVector(int     *, long, long);
void free_fMatrix(int    **, long, long, long, long);

double               /* OUT: maximum score in sum matrix                          */
/*******/
dyn_pr_n2
/*******/
/* Sequence 1 runs down   matrix index by row index i.
   Sequence 2 runs across matrix index by row index j. */
(double **s        /* IN:  similarity matrix               [0..rows-1][0..cols-1] */, 
 int *iOpnPen      /* IN:  gap-open pen. vector for seq. 1 [0..rows-1]            */, 
 int *iExtPen      /* IN:  gap-extend pen. vec. for seq. 1 [0..rows-1]            */, 
 int *jOpnPen      /* IN:  gap-open pen. vector for seq. 2 [0..cols-1]            */, 
 int *jExtPen      /* IN:  gap-extend pen. vec. for seq. 2 [0..cols-1]            */, 
 int rows          /* IN:  length of sequence 1 (Runs down matrix,  indexed by i) */,
 int cols          /* IN:  length of sequence 2 (Runs across matrix, indexed by j)*/,
 double **SS       /* OUT: sum matrix                      [0..rows-1][0..cols-1] */, 
 double **TrRow    /* OUT: traceback matrix with i index   [0..rows-1][0..cols-1] */, 
 double **TrCol    /* OUT: traceback matrix with j index   [0..rows-1][0..cols-1] */,
 int *iOverallMax  /* OUT: i index where overall max occurs 0..rows-1             */,
 int *jOverallMax  /* OUT: j index where overall max occurs 0..cols-1             */
)
{
  double TrialScore, ColMax, BlkMax, RowMax;
  int 
    i,j,i2,j2, jRowMax,
    iBlkMax, jBlkMax, 
    iGapLength, jGapLength, iGapPenalty, jGapPenalty;
  double OverallMax = - INF_INT;
  int iColMax[MAX_COLS];
  int Parm_FastBlkMax = 0;
  int Parm_SlowBlkMax = 0;
  
  for(i=0; i< rows; i++) {
    SS[i][0] = s[i][0];
    TrRow[i][0] = STOP_TRACEBACK ;
    TrCol[i][0] = STOP_TRACEBACK ;
  }
  for(j=0; j< cols; j++) {
    SS[0][j] = s[0][j];
    TrRow[0][j] = STOP_TRACEBACK ;
    TrCol[0][j] = STOP_TRACEBACK ;
    iColMax[j] = 0;
  }
  
  for(i=1; i< rows; i++) {

    jRowMax = 0;
    jBlkMax = 0;
    iBlkMax = iColMax[0];
    
    for(j=1; j< cols; j++) {
      
      /* You are at position i,j, consider four
         possible maximum values: */
      
      /* 1. The value at SS[i-1][j-1] with no gaps */

      /* 2. The maximum on column j-1 up to row i-2. */

      for (ColMax = -INF_INT,i2=iColMax[j-1];
	   i2<i-1; i2++) {
	iGapLength = i - i2 - 1;
	iGapPenalty = iOpnPen[i] + (iGapLength - 1) * iExtPen[i];
	TrialScore = SS[i2][j-1] - iGapPenalty ;
	if (TrialScore >= ColMax) {
	  iColMax[j-1] = i2 ;
	  ColMax = TrialScore ;
	}
      }
      
      /* 3. The maximum on row i-1 up to column j-2. */

      RowMax = -INF_INT ;
      for (j2=jRowMax; j2<j-1; j2++) {
	jGapLength = j - j2 - 1;
	jGapPenalty = jOpnPen[j] + (jGapLength - 1) * jExtPen[j];
	TrialScore = SS[i-1][j2] - jGapPenalty ;
	if (TrialScore >= RowMax) {
	  jRowMax = j2 ;
	  RowMax = TrialScore ;
	}
      }
      
      /* 4. The maximum in block 
	 with bottom right-hand corner at (i-2,j-1) 
	 and top left-hand corner at (iBlkMax, jBlkMax). */
 
      if (Parm_SlowBlkMax) {

	BlkMax = -INF_INT ;
	for (i2=iBlkMax; i2 < i-1; i2++) {
	  for (j2=jBlkMax; j2 < j-1; j2++) {
	    jGapLength = j - j2 - 1;
	    jGapPenalty = jOpnPen[j] + (jGapLength - 1) * jExtPen[j];
	    iGapLength = i - i2 - 1;
	    iGapPenalty = iOpnPen[i] + (iGapLength - 1) * iExtPen[i];
	    TrialScore = SS[i2][j2] - iGapPenalty - jGapPenalty;
	    if (TrialScore >= BlkMax) {
	      jBlkMax = j2 ;
	      iBlkMax = i2 ;
	      BlkMax = TrialScore ;
	    }
	  }
	}
	
      } else if (Parm_FastBlkMax) {
	
	BlkMax = -INF_INT ;
	for (j2=jBlkMax; j2 < j-1; j2++) {
	  i2 = iColMax[j2];
	  if (i2 < i - 1) {
	    jGapLength = j - j2 - 1;
	    jGapPenalty = jOpnPen[j] + (jGapLength - 1) * jExtPen[j];
	    iGapLength = i - i2 - 1;
	    iGapPenalty = iOpnPen[i] + (iGapLength - 1) * iExtPen[i];
	    TrialScore = SS[i2][j2] - iGapPenalty - jGapPenalty;
	    if (TrialScore >= BlkMax) {
	      jBlkMax = j2 ;
	      iBlkMax = i2 ;
	      BlkMax = TrialScore ;
	    }
	  }
	}
	
      } else {
	jBlkMax = j2 ;
	iBlkMax = i2 ;
	BlkMax = -INF_INT ;
      }
	
      
      if /* Case 1 */ 
	(SS[i-1][j-1] >= ColMax &&
	 SS[i-1][j-1] >= RowMax &&
	 SS[i-1][j-1] >= BlkMax) {
	  TrRow[i][j] =  i-1 ;
	  TrCol[i][j] =  j-1;
	  SS[i][j] = s[i][j] + SS[i-1][j-1] ;
      } else if /* Case 2 */
	(/* ColMax >= SS[i-1][j-1] .. True anyway! */
	 ColMax >= RowMax &&
	 ColMax >= BlkMax) {
	  TrRow[i][j] =  iColMax[j-1] ;
	  TrCol[i][j] =  j-1;
	  SS[i][j] = s[i][j] + ColMax ;
      } else if /* Case 3 */
	(/* RowMax >= SS[i-1][j-1]  ..True anyway! */
	 /* RowMax >= ColMax        ..True anyway! */
	 RowMax >= BlkMax) {
	  TrRow[i][j] =  i-1;
	  TrCol[i][j] =  jRowMax ;
	  SS[i][j] = s[i][j] + RowMax ;
      } else if /* Case 4 */
	(BlkMax >= SS[i-1][j-1] && /*True anyway! */
	 BlkMax >= ColMax &&       /*True anyway! */
	 BlkMax >= RowMax) {       /*True anyway! */
	  TrRow[i][j] =  iBlkMax;
	  TrCol[i][j] =  jRowMax ;
	  SS[i][j] = s[i][j] + BlkMax ;
      } else {
	printf("Error somehow not in case 1 to 4.\n");
      }
      
      if (SS[i][j] > OverallMax) {
	OverallMax = SS[i][j];
	*iOverallMax = i ;
	*jOverallMax = j ;
      }
    } /* j-loop */
  } /* i-loop */
  return OverallMax ;
}

int                /* OUT: number of equivalenced residues = n                       */
/*--------------*/
TrMat2Equivalences 
/*--------------*/
(
 int rows          /* IN:  length of sequence 1                                      */,
 int cols          /* IN:  length of sequence 2                                      */,
 double **TrRow    /* IN:  traceback matrix with row index   [0..rows-1][0..cols-1]  */, 
 double **TrCol    /* IN:  traceback matrix with col index   [0..rows-1][0..cols-1]  */,
 int iOverallMax   /* IN:  row index where overall max occurs 0..rows-1              */,
 int jOverallMax   /* IN:  col index where overall max occurs 0..cols-1              */,
 int *equiv1       /* OUT: seq. 1 equivalenced residues, using fortran conv. [0..n-1]*/,
 int *equiv2       /* OUT: seq. 2 equivalenced residues, using fortran conv. [0..n-1]*/,
 int *nbrks        /* OUT: number of chain breaks                                    */
)
{
  int ReversedEquiv1[MAX_ROWS];
  int ReversedEquiv2[MAX_COLS];
  int n = 0; 
  int i = iOverallMax ;
  int j = jOverallMax ;
  int ii, jj,nn;
  do {
    ReversedEquiv1[n] = i ;
    ReversedEquiv2[n] = j ;
    n++ ;
    ii = (int)TrRow[i][j] ;
    jj = (int)TrCol[i][j] ;
    if (ii != STOP_TRACEBACK && 
	!(ii == i - 1 && jj == j - 1)) {
      *nbrks = *nbrks + 1 ;
    }
    i = ii; j = jj;
  } while (i != STOP_TRACEBACK);
  
  for(i=0,nn=n; nn>0;i++,nn--) {
    /* Plus 1 changes from C to fortran array conventions ! */
    equiv1[i] = ReversedEquiv1[nn-1] + 1;
    equiv2[i] = ReversedEquiv2[nn-1] + 1; 
  }

  return n ;
}    