#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define ND 5
#define N  3
#define Pi 3.1415926535897932384626433832795
#define ee 2.7182818284590452353602874713527

int    *iVector(long, long);
float  *fVector(long, long);
double *dVector(long, long);
float  **fMatrix(long, long, long, long);
double **dMatrix(long, long, long, long);
void free_iVector(int     *, long, long);
void free_fVector(float   *, long, long);
void free_dVector(double  *, long, long);
void free_dMatrix(double **, long, long, long, long);
void free_fMatrix(float  **, long, long, long, long);

double mvmult( int, int, double **, double *, double *);
double mmmult( int, int, double **, double **, double **);
double transpose( int, double **, double **);

double igamc(double, double), igam(double, double);
double maxoffd(int, double *);
double mtransp(int, double *, double *);
double tritosquare(int, double [], double []);      
double mvmpy( int, int, double *, double *, double *); 

void eigens( double[], double[], double[], int );

double AddLog(double a, double b);

double hmmlogs(double **xmod, double **xque, int lenm, int lenq, char *hmmID)
{
int i, j, k, l;
int MKp1 = 1;
int IK   = 2;
int DKp1 = 3;
int LastState;
int LastSymbol;
float fp1, fp2, fp3, fp4, fp5, fp6, fp7, fp8, fp9;
double tmpSum, x, y, z, scoreX, scoreY, scoreZ;
double LTotalProb, TotalProb = -1;
double BlackHole = -pow(-10.0,6.0);
double **Lfm, **Lfi, **Lfd, **Lem, **Lei, **Lai, **Lam, **Lad;
double **ei, **em, **ai, **am, **ad, **distem;
double **Matr1, **Matr2, **RotMatr, *Vec, *RotVec;
double *DivX, *DivY, *DivZ, *RDistX, *RDistY, *RDistZ;
static double A[N*N], A0[N*N], X[N*N], D[N*N];
static double RR[N*N];
static double B[N], Y[N];
static int IPS[N];
int NumObsSymbols = lenq; /* 1;	lenq; */
int NumSymbols	  = lenq; /* 3;	lenq; */
int NumStates	  = lenm; /* 2;	lenm; */
FILE *pFile;

Lfm = dMatrix(0,NumStates,0,NumObsSymbols);
Lfi = dMatrix(0,NumStates,0,NumObsSymbols);
Lfd = dMatrix(0,NumStates,0,NumObsSymbols);
ei  = dMatrix(0,NumStates,1,NumSymbols);
em  = dMatrix(0,NumStates,1,NumSymbols);
ai  = dMatrix(0,NumStates,1,3);
am  = dMatrix(0,NumStates,1,3);
ad  = dMatrix(0,NumStates,1,3);
Lei = dMatrix(0,NumStates,1,NumSymbols);
Lem = dMatrix(0,NumStates,1,NumSymbols);
Lai = dMatrix(0,NumStates,1,3);
Lam = dMatrix(0,NumStates,1,3);
Lad = dMatrix(0,NumStates,1,3);

if(!(pFile = fopen(hmmID, "r"))) { printf("Couldn't open file %s \n", hmmID);
	exit(1); 
}
else {
    fscanf(pFile, "%i\n", &lenm);
	for(i=0;i<=lenm;i++) {   /* there should be lenm+1 lines with probabilities */
		fscanf(pFile, "%f %f %f %f %f %f %f %f %f\n", 
				&fp1, &fp2, &fp3, &fp4, &fp5, &fp6, &fp7, &fp8, &fp9);
		am[i][1]=fp1; am[i][2]=fp2; am[i][3]=fp3;
		ai[i][1]=fp4; ai[i][2]=fp5; ai[i][3]=fp6; 
		ad[i][1]=fp7; ad[i][2]=fp8; ad[i][3]=fp9;
	} 
	fclose(pFile);
}

distem	= dMatrix(0,lenm,1,lenq);
Matr1	= dMatrix(1,N,1,N);
Matr2	= dMatrix(1,N,1,N);
RotMatr	= dMatrix(1,N,1,N);
Vec		= dVector(1,N);
RotVec	= dVector(1,N);

DivX	= dVector(0,ND);
DivY	= dVector(0,ND);
DivZ	= dVector(0,ND);
RDistX	= dVector(0,ND);
RDistY	= dVector(0,ND);
RDistZ	= dVector(0,ND);

for(k=0; k<ND; k++) {
	DivX[k] = igam(0.50e0, 0.50e0*(k+1)*(k+1)) - igam(0.50e0, 0.50e0*(k*k));
	DivY[k] = igam(0.50e0, 0.50e0*(k+1)*(k+1)) - igam(0.50e0, 0.50e0*(k*k));
	DivZ[k] = igam(0.50e0, 0.50e0*(k+1)*(k+1)) - igam(0.50e0, 0.50e0*(k*k));
} 

DivX[ND] = 1.0e0 - igam(0.50e0, 0.50e0*(ND*ND));
DivY[ND] = 1.0e0 - igam(0.50e0, 0.50e0*(ND*ND));
DivZ[ND] = 1.0e0 - igam(0.50e0, 0.50e0*(ND*ND));

/* printf("\n The histogram-score vector:\n" ); */
/* for(k=0; k<=ND; k++) printf("%f\n", DivX[k]); */

for(i=1;i<=lenm;i++) {
    /* Initialize the triangular storage vector A */
	A[0] = xmod[i][1+3]; A[1] = xmod[i][4+3]; A[2] = xmod[i][2+3];
	A[3] = xmod[i][5+3]; A[4] = xmod[i][6+3]; A[5] = xmod[i][3+3];
	/* Unpack A into a symmetric matrix A0 and solve the eigenvalue problem */
	tritosquare( N, A, A0 );
	eigens( A, RR, B, N );
	mtransp( N, RR, RR );
    /* Form the square NxN Rotation matrix: */
	for(k=0;k<N;k++) {
		for(l=0;l<N;l++) { Matr1[k+1][l+1] = RR[3*k+l]; } /* l */
	} 
	transpose( N, Matr1, RotMatr );				/* Transposed Rotation NxN matrix */
 
	B[0] = fabs(B[0]); B[1] = fabs(B[1]); B[2] = fabs(B[2]);

/*	printf("%f %f %f\n",B[0],B[1],B[2]); */
/*	B[0] = B[0]*1.0; B[1] = B[1]*1.0; B[2] = B[2]*1.0;  */

	if(B[0] < 1.0e-3) {B[0] = 0.001;}
	if(B[1] < 1.0e-3) {B[1] = 0.001;}
	if(B[2] < 1.0e-3) {B[2] = 0.001;}

	for(k=0;k<=ND;k++) {
		RDistX[k] = sqrt( B[0]*(double)k );
		RDistY[k] = sqrt( B[1]*(double)k );
		RDistZ[k] = sqrt( B[2]*(double)k );
	} 

	for(j=1;j<=lenq;j++) {
		Vec[1] = xque[j][1]-xmod[i][1];			/* Need to translate Vec->Vec' */
		Vec[2] = xque[j][2]-xmod[i][2];			/* to the origin of the i-th ellipoid  */
		Vec[3] = xque[j][3]-xmod[i][3];

		mvmult( N, N, RotMatr, Vec, RotVec );	/* Coordinates of Vec' are rotated */

		x=RotVec[1]; y=RotVec[2]; z=RotVec[3];

/*		scoreX = 1.0e-50; scoreY = 1.0e-50; scoreZ = 1.0e-50; */ /* Default values for scores */
/* Severe underflow problems occur in igamc unless we restrict the arg value: */

		scoreX = DivX[ND];	
		scoreY = DivY[ND];	
		scoreZ = DivZ[ND];
		for(k=0; k<ND; k++) {  /* Find out where the Query atom j is located on the histogram */
			if(fabs(x) >= RDistX[k] && fabs(x) < RDistX[k+1]) scoreX = DivX[k];
			if(fabs(y) >= RDistY[k] && fabs(y) < RDistY[k+1]) scoreY = DivY[k];
			if(fabs(z) >= RDistZ[k] && fabs(z) < RDistZ[k+1]) scoreZ = DivZ[k];
		} 

/*		if( x*x/(2.0*B[0]) <= 100 ) scoreX = igamc(0.5, x*x/(2.0*B[0])); */
/*		if( y*y/(2.0*B[1]) <= 100 ) scoreY = igamc(0.5, y*y/(2.0*B[1])); */
/*		if( z*z/(2.0*B[2]) <= 100 ) scoreZ = igamc(0.5, z*z/(2.0*B[2])); */

		em[i][j] = scoreX*scoreY*scoreZ;

		ei[i][j] =	1.0;
		distem[i][j] = sqrt(pow((xmod[i][1]-xque[j][1]),2)+
							pow((xmod[i][2]-xque[j][2]),2)+
							pow((xmod[i][3]-xque[j][3]),2) );  /* test with Vec! */
	} /* j */
} /* i */

for(j=1;j<=lenq;j++) {
	ei[0][j] =	1.0;
	em[0][j] =	0.0;
	distem[0][j] =	0.0;
}
/*
printf("\n 4x4 left upper corner of the total distance matrix:\n" );
for(i=0; i<=3; i++) {
    for(j=1; j<=4; j++) {
		printf("%f  ", distem[i][j]);
	}
	printf("\n");
}

printf("\n 4x9 left upper corner of the atom-emission matrix:\n" );
for(i=0; i<=5; i++) {
    for(j=1; j<=8; j++) {
		printf("%f  ", em[i][j]);
	}
	printf("\n");
}
*/

/*
printf("\n 4x10 bottom part of the corextract matrix (inverse order):\n" );
for(i=1; i<=4; i++) {
    for(j=0; j<=9; j++) {
		printf("%f  ", xmod[lenm-i+1][j]);
	}
	printf("\n");
}

printf("\n 4x9 bottom part of the hmmparams matrix (inverse order):\n" );
for(i=0; i<=3; i++) {
    for(j=1; j<=3; j++) { printf("%f ", am[NumStates-i][j]); }
    for(j=1; j<=3; j++) { printf("%f ", ai[NumStates-i][j]); }
    for(j=1; j<=3; j++) { printf("%f ", ad[NumStates-i][j]); }
	printf("\n");
}
*/
/* strcpy(hmmID, "Try to change it."); */ /* Now works fine. */

for(i=0; i<=NumStates; i++) {
	for(j=1;j<=3;j++) {
		if(ai[i][j] != 0) { Lai[i][j] = log(ai[i][j]); }
			else { Lai[i][j] = BlackHole; }
		if(am[i][j] != 0) { Lam[i][j] = log(am[i][j]); }
			else { Lam[i][j] = BlackHole; }
		if(ad[i][j] != 0) { Lad[i][j] = log(ad[i][j]); }
			else { Lad[i][j] = BlackHole; }
	} /*j */
} /*k */

for(i=0; i<=NumStates; i++) {
	for(j=1;j<=NumSymbols;j++) {
		if(ei[i][j] != 0) { Lei[i][j] = log(ei[i][j]); }
			else { Lei[i][j] = BlackHole; }
		if(em[i][j] != 0) { Lem[i][j] = log(em[i][j]); }
			else { Lem[i][j] = BlackHole; }
	} /* j */
} /* i */

for(i=0;i<=NumStates;i++) {
	for(j=0;j<=NumObsSymbols;j++) {
		Lfm[i][j] = BlackHole;
		Lfi[i][j] = BlackHole;
		Lfd[i][j] = BlackHole;
	} /* j */
}/* i */

Lfm[0][0] = 0.0;
Lfd[1][0] = log(am[0][DKp1]);

if(NumStates > 1) { 
	for(k=2; k<=NumStates; k++) {
		tmpSum = 0.0;
		for(j=1;j<=k-1;j++) {
		tmpSum = tmpSum + Lad[j][DKp1]; 
		} /*j*/
		Lfd[k][0] = Lam[0][DKp1] + tmpSum;
	}/*k */
} /*if*/

LastState  = NumStates;
LastSymbol = NumObsSymbols;

/*for(i=0; i<=NumStates; i++) {
   printf("%f %f\n", Lfd[i][0], Lfd[i][1] );
}
*/

for(i=1; i<=NumObsSymbols; i++) {

    Lfi[0][i] = Lei[0][i] + 
	AddLog(Lfm[0][i-1] + Lam[0][IK], AddLog(Lfi[0][i-1] + Lai[0][IK], Lfd[0][i-1] + Lad[0][IK]));

    for(k=1; k <= NumStates; k++) {

	Lfi[k][i] = Lei[k][i] + 
		AddLog(Lfm[k][i-1] +
		Lam[k][IK],AddLog(Lfi[k][i-1] + Lai[k][IK],Lfd[k][i-1] + Lad[k][IK]));

    Lfm[k][i] = Lem[k][i] + 
		AddLog(Lfm[k-1][i-1]+Lam[k-1][MKp1],AddLog(Lfi[k-1][i-1]+Lai[k-1][MKp1],
				Lfd[k-1][i-1]+Lad[k-1][MKp1]));

    Lfd[k][i] = AddLog(Lfm[k-1][i]+Lam[k-1][DKp1],AddLog(Lfi[k-1][i]+Lai[k-1][DKp1],
				Lfd[k-1][i] + Lad[k-1][DKp1]));
	}
}

/* for(i=0; i<=NumStates; i++) {
    printf("%f %f\n", Lfi[i][0], Lfi[i][1] );
}
*/

LTotalProb = AddLog(Lfm[LastState][LastSymbol] + Lam[LastState][MKp1], 
			 AddLog(Lfi[LastState][LastSymbol] + Lai[LastState][MKp1], 
			 Lfd[LastState][LastSymbol] + Lad[LastState][MKp1]));

/*TotalProb = exp(LTotalProb);*/

free_dMatrix(Lfi,0,NumStates,0,NumObsSymbols);
free_dMatrix(Lfm,0,NumStates,0,NumObsSymbols);
free_dMatrix(Lfd,0,NumStates,0,NumObsSymbols);
free_dMatrix(ei,0,NumStates,1,NumSymbols);
free_dMatrix(em,0,NumStates,1,NumSymbols);
free_dMatrix(ai,0,NumStates,1,3);
free_dMatrix(am,0,NumStates,1,3);
free_dMatrix(ad,0,NumStates,1,3);
free_dMatrix(Lei,0,NumStates,1,NumSymbols);
free_dMatrix(Lem,0,NumStates,1,NumSymbols);
free_dMatrix(Lai,0,NumStates,1,3);
free_dMatrix(Lam,0,NumStates,1,3);
free_dMatrix(Lad,0,NumStates,1,3);
free_dMatrix(distem,0,lenm,1,lenq);
free_dMatrix(RotMatr,1,N,1,N);
free_dMatrix(Matr1,1,N,1,N);
free_dMatrix(Matr2,1,N,1,N);
free_dVector(Vec,1,N);
free_dVector(RotVec,1,N);
free_dVector(DivX,0,ND);
free_dVector(DivY,0,ND);
free_dVector(DivZ,0,ND);
free_dVector(RDistX,0,ND);
free_dVector(RDistY,0,ND);
free_dVector(RDistZ,0,ND);

return -LTotalProb;
} /* hmmlogs */


double AddLog(double a, double b)
{
if(fabs(b-a) > 1000) {
	if(fabs(b) > fabs(a)) {return a;}
	else {return b;}
}
else { return a + log(1.0 + pow(ee,b - a)); }
} /* AddLog*/

/*  HMM scoring example:
ei[0][1]=0.20; ei[0][2]=0.40; ei[0][3]=0.40; 
ei[1][1]=0.30; ei[1][2]=0.35; ei[1][3]=0.35; 
ei[2][1]=0.40; ei[2][2]=0.30; ei[2][3]=0.30; 

em[0][1]=0.00; em[0][2]=0.00; em[0][3]=0.00; 
em[1][1]=0.10; em[1][2]=0.45; em[1][3]=0.45; 
em[2][1]=0.50; em[2][2]=0.25; em[2][3]=0.25; 

ai[0][1]=0.60; ai[0][2]=0.11; ai[0][3]=0.29; 
ai[1][1]=0.80; ai[1][2]=0.12; ai[1][3]=0.08; 
ai[2][1]=0.87; ai[2][2]=0.13; ai[2][3]=0.00; 

am[0][1]=0.50; am[0][2]=0.30; am[0][3]=0.20; 
am[1][1]=0.75; am[1][2]=0.10; am[1][3]=0.15; 
am[2][1]=0.74; am[2][2]=0.26; am[2][3]=0.00; 

ad[0][1]=0.00; ad[0][2]=0.00; ad[0][3]=0.00; 
ad[1][1]=0.01; ad[1][2]=0.03; ad[1][3]=0.96; 
ad[2][1]=0.56; ad[2][2]=0.44; ad[2][3]=0.00; 

*/
