#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "umath.h"

#define RELEASE char*

typedef double rotate_t[3][3];
typedef double xlate_t[3];
typedef struct Coord { double x, y, z; } CoordType;

int getrms(double *, double);
int nseq    = 2300;

double MAX2(double, double);
double MIN2(double, double);
double FormRotMatrX(double **, double);
double FormRotMatrY(double **, double);
double FormRotMatrZ(double **, double);
double AlignZ(double, double, double, double *, double *);
int FindMaxXYZDistIdxLast(double **, int);
int FindMaxXYDistIdxLast(double **, int);
double TRmoveXYZstruc(double *,double **,double **, int);
double TRvec(double *, double **, double *);
double findCM(double **, int, double *);
double blossom62(char, char);
char convert321(char *);
double distrms(CoordType *a, CoordType *b, int npts);
int check_eq(int *, int *, int *, int *, int);
double rotrans(double **, int, rotate_t, xlate_t);
double bestfit(rotate_t , xlate_t , CoordType *, CoordType *, const int );

int     *iVector(long, long);
int    **iMatrix(long, long, long, long);
double  *dVector(long, long);
char    *cVector(long, long);
double **dMatrix(long, long, long, long);
char   **cMatrix(long, long, long, long);
void free_iVector(int     *, long, long);
void free_iMatrix(int    **, long, long, long, long);
void free_dVector(double     *, long, long);
void free_dMatrix(double    **, long, long, long, long);
void free_cMatrix(char      **, long, long, long, long);
void free_cVector(char     *, long, long);

double dyn_pr_n2(double **, int *, int *, int *, int *, int, 
                 int, double **, double **, double **, int *, int *);
int TrMat2Equivalences(int, int, double **, double **, int, int, int *, int *, int *);
void show_alignment(char *, char *, int, int *, int *, char *, char *, char *);

int MAXSEQ = 1000;

int main(int argc, char* argv[])
{
 double ret_val, RMS, PID, tempval;
 double RMSmin, RMSLast;
 int nEqPairsLast = -1;
 int iSameRMSmin, iSameRMSLast;
 int i, j;
 int rows, cols;
 double ** s;
 double ** SS;
 int * iOpnPen;
 int * iExtPen;
 int * jOpnPen;
 int * jExtPen;
 double ** TrRow;
 double ** TrCol;
 int iOverallMax;
 int jOverallMax;

 int *iina, *iina_sav;
 int *jinb, *jinb_sav;
 int nbrks, nEqPairs;

 float fp1, fp2, fp3;
 char   ** aa_que, ** aa_tar;
 double ** x_que, **x_que_original;
 double ** x_tar;
 double ** x_que_sav;
 double * Tcm;
 FILE *stream;

 char *s1, *s2, *s3, *s4, *a_que, *a_tar, *csqa1, *csqa2, *csqam;
 int d1, d2;
 int StringMax = 4;

 rotate_t U;
 xlate_t  T;

 int stay, nEqPairs_sav;
 int iter = 0;
 char aa;

 double *TMPvec, *Tzero, *Tcm_que, *Tcm_tar; 
 double **Rtemp, **Runity;
 double angle_b, angle_c, angle_b_que, angle_c_que, angle_b_tar, angle_c_tar;
 double xdir_que, ydir_que, rdir_que, xdir_tar, ydir_tar, rdir_tar;
 int iMaxDist;

 int    getrms_val;
 double ret_hscore;
 int i1, k, imin = -1;

 CoordType *str1, *str2;

 int structal     = 0;    /* Flag to use OLD output similar to the one from STRUCTAL*/
 RMSmin = RMSLast = 100000;
 iSameRMSmin = iSameRMSLast = 0;

 if( argc == 2 && strcmp( argv[1], "structal" ) == 0 ) {
	 printf("LSQRMS was called with argument %s \n\n",argv[1]);
	 structal = 1;
 }

 /* printf("Program LSQ3DHMM. Version 1.01\n"); */
 stay = 1; nEqPairs_sav = -1; nEqPairs = -1;

 s1 = cVector(1,StringMax);
 s2 = cVector(1,StringMax);
 s3 = cVector(1,3);
 s4 = cVector(1,StringMax);

/* ATOM      2  CA  ASP H  19      30.313   5.590  12.619  */
    stream = fopen( "query.pdb", "r" );
	if( stream == NULL ) {
	   printf( "Couldn't open file query.pdb\n" );
	   exit(1); }
	else   {
    i=0;
	while( !feof( stream ) ) {
	  fscanf( stream, "%4c %d %2c %3c %c %d %f %f %f\n", s1, &d1, s2, s3, s4, &d2,
		    &fp1, &fp2, &fp3);
	  aa = convert321( s3 );
	  i++;
	} /* while */
    fclose( stream );
	}
	rows = i;
    stream = fopen( "query.pdb", "r" );
    x_que     = dMatrix(1,3,0,rows-1);
    x_que_sav = dMatrix(1,3,0,rows-1);
    x_que_original = dMatrix(1,3,0,rows-1);
    a_que     = cVector(0,rows-1);
    aa_que    = cMatrix(0,rows-1,0,2);
	for(i=0;i<rows;i++) {
	  fscanf( stream, "%4c %d %2c %3c %c %d %f %f %f\n", s1, &d1, s2, s3, s4, &d2,
		    &fp1, &fp2, &fp3);
	  aa = convert321( s3 );
	  a_que[i] = aa;
	  aa_que[i][0] = s3[0];
	  aa_que[i][1] = s3[1];
	  aa_que[i][2] = s3[2];
      x_que[1][i]=fp1; x_que_sav[1][i]=fp1; x_que_original[1][i]=fp1;
      x_que[2][i]=fp2; x_que_sav[2][i]=fp2; x_que_original[2][i]=fp2;
      x_que[3][i]=fp3; x_que_sav[3][i]=fp3; x_que_original[3][i]=fp3;
	} /* for(i) */
    fclose( stream );

    stream = fopen( "target.pdb", "r" );
	if( stream == NULL ) {
	   printf( "Couldn't open file target.pdb\n" );
	   exit(1); }
	else   {
    i=0;
	while( !feof( stream ) ) {
	  fscanf( stream, "%4c %d %2c %3c %c %d %f %f %f\n", s1, &d1, s2, s3, s4, &d2,
		    &fp1, &fp2, &fp3);
	  aa = convert321( s3 );
	  i++;
	} /* while */
    fclose( stream );
	}
	cols = i;
    stream = fopen( "target.pdb", "r" );
    x_tar     = dMatrix(1,3,0,cols-1);
    a_tar     = cVector(0,cols-1);
	aa_tar    = cMatrix(0,cols-1,0,2);
	for(i=0;i<cols;i++) {
	  fscanf( stream, "%4c %d %2c %3c %c %d %f %f %f\n", s1, &d1, s2, s3, s4, &d2,
		    &fp1, &fp2, &fp3);
	  aa = convert321( s3 );
	  a_tar[i] = aa;
	  aa_tar[i][0] = s3[0];
	  aa_tar[i][1] = s3[1];
	  aa_tar[i][2] = s3[2];
      x_tar[1][i]=fp1; 
      x_tar[2][i]=fp2; 
      x_tar[3][i]=fp3; 
	} /* for(i) */
    fclose( stream );
/* ATOM     10  CA  GLU H  20      30.998   8.719  10.360  */

 iOverallMax = -1;
 jOverallMax = -1;

 s       = dMatrix(0,rows-1,0,cols-1);
 SS      = dMatrix(0,rows-1,0,cols-1);
 TrRow   = dMatrix(0,rows-1,0,cols-1);
 TrCol   = dMatrix(0,rows-1,0,cols-1);
 iOpnPen = iVector(0,rows-1);
 iExtPen = iVector(0,rows-1);
 jOpnPen = iVector(0,cols-1);
 jExtPen = iVector(0,cols-1);

 iina_sav = iVector(0,MAXSEQ);
 jinb_sav = iVector(0,MAXSEQ);

/* FIRST OF ALL, MOVE que->CM(tar); NEXT, align PRINCIPAL AXES OF INERTIA: */

 Tcm     = dVector(1,3); 
 Tcm_que = dVector(1,3); 
 Tcm_tar = dVector(1,3); 
 Tzero   = dVector(1,3); 
 TMPvec  = dVector(1,3); 
 Rtemp   = dMatrix(1,3,1,3);
 Runity  = dMatrix(1,3,1,3);
 Rtemp[1][1]  = 1.0; Rtemp[1][2]  = 0.0; Rtemp[1][3]  = 0.0;
 Rtemp[2][1]  = 0.0; Rtemp[2][2]  = 1.0; Rtemp[2][3]  = 0.0;
 Rtemp[3][1]  = 0.0; Rtemp[3][2]  = 0.0; Rtemp[3][3]  = 1.0;
 Runity[1][1] = 1.0; Runity[1][2] = 0.0; Runity[1][3] = 0.0;
 Runity[2][1] = 0.0; Runity[2][2] = 1.0; Runity[2][3] = 0.0;
 Runity[3][1] = 0.0; Runity[3][2] = 0.0; Runity[3][3] = 1.0;
 TMPvec[1]  = 2; TMPvec[2]  = 3;   TMPvec[3]  = 4;
 Tzero[1]   = 00.e0; Tzero[2] = 0.0e0; Tzero[3]   = 0.0e0;

 ret_val = findCM(x_tar,cols,Tcm);

 Tcm_tar[1] = Tcm[1];  Tcm_tar[2] = Tcm[2];  Tcm_tar[3] = Tcm[3]; /* saved inv translation */
 Tcm[1] = -Tcm[1];  Tcm[2] = -Tcm[2];  Tcm[3] = -Tcm[3];

 ret_val = TRmoveXYZstruc(Tcm,Runity,x_tar,cols);

 ret_val = findCM(x_tar,cols,Tcm);

 iMaxDist = FindMaxXYZDistIdxLast(x_tar,cols);
 ret_val  = AlignZ(x_tar[1][iMaxDist],x_tar[2][iMaxDist],x_tar[3][iMaxDist],&angle_b,&angle_c);
 angle_b_tar = -angle_b; angle_c_tar = -angle_c; /* saved inv rotation */

 ret_val = FormRotMatrY(Rtemp,angle_b);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
 ret_val = FormRotMatrX(Rtemp,angle_c);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);

 iMaxDist = FindMaxXYDistIdxLast(x_tar,cols);

 rdir_tar = sqrt( SQR(x_tar[1][iMaxDist]) + SQR(x_tar[2][iMaxDist]) );
 xdir_tar = x_tar[1][iMaxDist]/rdir_tar; /* <- need it to determine XY-plane rotation */
 ydir_tar = x_tar[2][iMaxDist]/rdir_tar; /* <- need it to determine XY-plane rotation */

/* Do the same for the QUERY: */
 ret_val = findCM(x_que,rows,Tcm);
 Tcm_que[1] = Tcm[1];  Tcm_que[2] = Tcm[2];  Tcm_que[3] = Tcm[3]; /* saved inv translation */

 Tcm[1] = -Tcm[1];  Tcm[2] = -Tcm[2];  Tcm[3] = -Tcm[3];

 ret_val = TRmoveXYZstruc(Tcm,Runity,x_que,rows);

 ret_val = findCM(x_que,rows,Tcm);

 iMaxDist = FindMaxXYZDistIdxLast(x_que,rows);
 ret_val  = AlignZ(x_que[1][iMaxDist],x_que[2][iMaxDist],x_que[3][iMaxDist],&angle_b,&angle_c);
 angle_b_que = -angle_b; angle_c_que = -angle_c; /* saved inv rotation */

 ret_val = FormRotMatrY(Rtemp,angle_b);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);
 ret_val = FormRotMatrX(Rtemp,angle_c);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);

 iMaxDist = FindMaxXYDistIdxLast(x_que,rows);

 rdir_que = sqrt( SQR(x_que[1][iMaxDist]) + SQR(x_que[2][iMaxDist]) );
 xdir_que = x_que[1][iMaxDist]/rdir_que; /* <- need it to determine XY-plane rotation */
 ydir_que = x_que[2][iMaxDist]/rdir_que; /* <- need it to determine XY-plane rotation */

 angle_b = acos(xdir_tar);
 angle_c = acos(xdir_que);

 ret_val = FormRotMatrZ(Rtemp,angle_c-angle_b);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_que,rows);

/* CHECK COORDINATES: */
/*
 stream = fopen( "coordchk.txt", "w" );
 if( stream == NULL )
   printf( "Couldn't open file coordchk.txt\n" );
 else   {
   fprintf( stream, "Query: \n");
   for(i=0;i<rows;i++) {
	  fprintf( stream, "%12.3f %8.3f %8.3f\n", x_que[1][i], x_que[2][i], x_que[3][i]);
	}
   fprintf( stream, "Target: \n");
   for(i=0;i<cols;i++) {
	  fprintf( stream, "%12.3f %8.3f %8.3f\n", x_tar[1][i], x_tar[2][i], x_tar[3][i]);
   }
   fclose( stream );
 }
*/
/* INITIAL STARTING POINT: SEQUENCE ALIGNMENT. Decide if needed (based on PID) */
 iina = iVector(0,MAXSEQ);
 jinb = iVector(0,MAXSEQ);
 for(i=0;i<=MAXSEQ;i++) {
   iina[i] = -1; jinb[i] = -1;
 }
 for(j=0;j<=cols-1;j++) {
     jOpnPen[j] = 0; jExtPen[j] = 0;
 }
 for(i=0;i<=rows-1;i++) {
	 iOpnPen[i] = 0;  iExtPen[i] = 0;
	for(j=0;j<=cols-1;j++) {       
		SS[i][j]    = 0.0;
        TrRow[i][j] = 0.0;
		TrCol[i][j] = 0.0;
	 }
 }
 /* Use substitution matrix over here: */
 for(i=0;i<rows;i++) {
	for(j=0;j<cols;j++) {
		if( a_que[i] == a_tar[j] ) s[i][j] = 10.0; /* primitive! */
		else s[i][j] = 0.0;                       /* primitive! */
/* blossom62 enhancement, not always good */
        tempval = blossom62(a_que[i],a_tar[j]);   
		s[i][j] += tempval;

	}
 }
 ret_val = dyn_pr_n2(s, iOpnPen, iExtPen, jOpnPen, jExtPen, rows, cols,
                     SS,  TrRow,   TrCol, &iOverallMax, &jOverallMax); 
 nbrks = -1;
// nEqPairs_sav = nEqPairs;
 nEqPairs = TrMat2Equivalences (rows,cols,TrRow,TrCol,
	                            iOverallMax,jOverallMax,iina,jinb,&nbrks);

 for(i=0;i<=MAXSEQ;i++) {
   iina_sav[i] = iina[i];
   jinb_sav[i] = jinb[i];
 }

 tempval = MIN2( (double)rows, (double)cols);
 PID = (double)nEqPairs/( tempval );

 if(PID > 0.1) { /* It makes sense to start with eq.pairs from sequence alignment: */

 str1 = (CoordType *) malloc(nEqPairs * sizeof(CoordType));
 str2 = (CoordType *) malloc(nEqPairs * sizeof(CoordType));

 for(i=0;i<nEqPairs;i++) {
   str1[i].x = x_que[1][iina[i]-1]; str2[i].x = x_tar[1][jinb[i]-1];
   str1[i].y = x_que[2][iina[i]-1]; str2[i].y = x_tar[2][jinb[i]-1];
   str1[i].z = x_que[3][iina[i]-1]; str2[i].z = x_tar[3][jinb[i]-1];		
 }

 RMS = bestfit(U, T, str1, str2, nEqPairs); 
 /*
 printf("PID, RMS, eq.pairs (from sequence alignment) = %f %f %d\n",PID*100,RMS,nEqPairs);
*/
 ret_val = rotrans(x_que, rows, U, T);

 if( RMS <= RMSmin ) {
   RMSmin = RMS;
   nEqPairs_sav = nEqPairs;
   for(i=0; i<rows; i++) {
     x_que_sav[1][i] = x_que[1][i]; 
     x_que_sav[2][i] = x_que[2][i]; 
     x_que_sav[3][i] = x_que[3][i]; 
   }
 }
 free( (RELEASE) (str1) );
 free( (RELEASE) (str2) );
 } else { /* PID < threshold */
   printf("PID is %f, rather low to start with SeqAli Eq. pairs\n",PID*100);
 }

 free_iVector(iina,0, MAXSEQ);
 free_iVector(jinb,0, MAXSEQ);

/* ##################### END OF THE INITIAL STARTING POINT ##################### */

 do{ iter++;    /* START OF MAIN ITERATION CYCLE: */
/* CHECK COORDINATES:
    stream = fopen( "coordchk.dat", "a" );
	if( stream == NULL )
	   printf( "Couldn't open file coordchk.dat\n" );
	else   {
    fprintf( stream, "Iteration # %d: \n",iter);
    fprintf( stream, "Query: \n");
	for(i=0;i<rows;i++) {
	  fprintf( stream, "%f %f %f\n", x_que[1][i], x_que[2][i], x_que[3][i]);
	}
    fprintf( stream, "Target: \n");
	for(i=0;i<cols;i++) {
	  fprintf( stream, "%f %f %f\n", x_tar[1][i], x_tar[2][i], x_tar[3][i]);
	}
    fclose( stream );
	}
*/
 iina     = iVector(0,MAXSEQ);
 jinb     = iVector(0,MAXSEQ);

 for(i=0;i<=MAXSEQ;i++) {
   iina[i] = -1;
   jinb[i] = -1;
 }

 for(i=0;i<=rows-1;i++) {
	for(j=0;j<=cols-1;j++) {
		s[i][j]     = sqrt( (x_que[1][i]-x_tar[1][j])*(x_que[1][i]-x_tar[1][j]) +
			                (x_que[2][i]-x_tar[2][j])*(x_que[2][i]-x_tar[2][j]) +
			                (x_que[3][i]-x_tar[3][j])*(x_que[3][i]-x_tar[3][j]) );
		s[i][j] = 20.0 / ( 1.0 + SQR(s[i][j]/5.0) );
		if( a_que[i] == a_tar[j] ) s[i][j] *= 1.2;
//		s[i][j] = s[i][j]*sqrt(sqrt((double)iter/10.0));
		SS[i][j]    = 0.0;
        TrRow[i][j] = 0.0;
		TrCol[i][j] = 0.0;
	 }
 }

 ret_val = dyn_pr_n2(s, iOpnPen, iExtPen, jOpnPen, jExtPen, rows, cols,
                     SS,  TrRow,   TrCol, &iOverallMax, &jOverallMax); 

 nbrks = -1;

// nEqPairs_sav = nEqPairs;
 nEqPairs = TrMat2Equivalences (rows,cols,TrRow,TrCol,
	                            iOverallMax,jOverallMax,iina,jinb,&nbrks);

 str1 = (CoordType *) malloc(nEqPairs * sizeof(CoordType));
 str2 = (CoordType *) malloc(nEqPairs * sizeof(CoordType));

 for(i=0;i<nEqPairs;i++) {
   str1[i].x = x_que[1][iina[i]-1]; str2[i].x = x_tar[1][jinb[i]-1];
   str1[i].y = x_que[2][iina[i]-1]; str2[i].y = x_tar[2][jinb[i]-1];
   str1[i].z = x_que[3][iina[i]-1]; str2[i].z = x_tar[3][jinb[i]-1];		
 }

 RMS = bestfit(U, T, str1, str2, nEqPairs); 

 /* printf("RMS, eq.pairs = %f %d\n",RMS,nEqPairs); */

 ret_val = rotrans(x_que, rows, U, T);
/*
 if( !(nEqPairs-nEqPairsLast) ) {
	 stay = check_eq(iina,jinb,iina_sav,jinb_sav,nEqPairs);
	 if( !stay ) {
	   if(RMS != RMSmin) stay = 1;
	 }
 }
*/

 if( RMS < RMSmin ) {
   RMSmin = RMS;
   nEqPairs_sav = nEqPairs;
   for(i=0; i<rows; i++) {
     x_que_sav[1][i] = x_que[1][i]; 
     x_que_sav[2][i] = x_que[2][i]; 
     x_que_sav[3][i] = x_que[3][i]; 
   }
   for(i=0;i<=MAXSEQ;i++) {
     iina_sav[i] = iina[i];
     jinb_sav[i] = jinb[i];
   }
 }

 if(RMS == RMSmin) {
   iSameRMSmin++;
   if( iSameRMSmin > 10) stay = 0;
 }
 else {
   if(RMS == RMSLast) { 
     iSameRMSLast++;
     if( iSameRMSLast > 10) stay = 0;
   }
 }

 RMSLast = RMS;
 nEqPairsLast = nEqPairs;

 free( (RELEASE) (str1) );
 free( (RELEASE) (str2) );

 free_iVector(iina,0, MAXSEQ);
 free_iVector(jinb,0, MAXSEQ);

/* printf("RMS = %f for %d equivalence pairs\n",RMS,nEqPairs); */

  if(iter > 100) stay = 0;
} while(stay); /* End of iteration cycle */ 

/* } while(iter<100); */
/* } while(stay && iter<100); */
/* } while( iter < 100); */

 /* printf("RMSmin = %f \n",RMSmin); */

/* Performing back-transformation of BOTH structures (QUERY will still be PMI-aligned!): */
 ret_val = FormRotMatrX(Rtemp,angle_c_tar);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_que_sav,rows);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
 ret_val = FormRotMatrY(Rtemp,angle_b_tar); 
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_que_sav,rows);
 ret_val = TRmoveXYZstruc(Tzero,Rtemp,x_tar,cols);
 ret_val = TRmoveXYZstruc(Tcm_tar,Runity,x_que_sav,rows);
 ret_val = TRmoveXYZstruc(Tcm_tar,Runity,x_tar,cols);

/* CHECK COORDINATES: */
/*
 stream = fopen( "rot_query.dat", "w" );
 if( stream == NULL )
   printf( "Couldn't open file rot_query.dat\n" );
 else {
   for(i=0;i<rows;i++) {
	  fprintf( stream, "%4s%7d%4s  %c%c%c %1s%4d%12.3f %8.3f %8.3f\n", "ATOM",i+1,"CA",
		  aa_que[i][0], aa_que[i][1], aa_que[i][2],"L",i+1,
		  x_que_sav[1][i],  x_que_sav[2][i], x_que_sav[3][i]);
   }
   fclose( stream );
 }
*/
 stream = fopen( "que.dat", "w" );
 if( stream == NULL )
   printf( "Couldn't open file que.dat\n" );
 else {
   fprintf( stream, "%d\n", rows);
   for(i=0;i<rows;i++) { 
      fprintf( stream, "%8.3f %8.3f %8.3f\n", x_que_sav[1][i], x_que_sav[2][i], x_que_sav[3][i]);
   }
   fclose( stream );
 }


 if(structal) {  /* STRUCTAL-LIKE OUTPUT REQUESTED */

   /* SHOW SEQUENCE ALIGNMENT BASED ON THE SAVED EQUIVALENCE PAIRS */
   csqa1 = cVector(0,2*MAXSEQ-1); 
   csqa2 = cVector(0,2*MAXSEQ-1); 
   csqam = cVector(0,2*MAXSEQ-1);

   for(i=0;i<2*MAXSEQ;i++) { csqa1[i] = '\0'; csqa2[i] = '\0'; csqam[i] = '\0'; }
   show_alignment(a_que, a_tar, nEqPairs_sav, iina_sav, jinb_sav, csqa1, csqa2, csqam);
   i=0; k=0;
   do{ if( csqa1[i] > 0 && csqa1[i] != '\0') k++;
     i++;
   } while ( csqa1[i] != '\0' );
   i1 = k;
   i=0; k=0;
   do{ if( csqa2[i] > 0 && csqa2[i] != '\0') k++;
     i++;
   } while ( csqa2[i] != '\0' );

   imin = (int)MIN2( (double)i1, (double)k);

   stream = fopen( "structal.out", "w" );
   if( stream == NULL )
     printf( "Couldn't open file structal.out\n" );
   else {
     fprintf( stream, "DUMP rmsnoe   %f (RMS deviation, before elim. bad)\n", RMSmin);
     fprintf( stream, "DUMP rmsbes   %f (RMS deviation, NO elim. bad in this version!)\n", RMSmin);

    fprintf( stream, " QUERY    > " );
    for(i=0;i<rows;i++) { fprintf( stream, "%c", a_que[i]); }
	fprintf( stream, "\n");

    fprintf( stream, " TARGET   > " );
    for(i=0;i<cols;i++) { fprintf( stream, "%c", a_tar[i]); }
	fprintf( stream, "\n");

     fprintf( stream, " query    > " );

   i=0; k=0;
   do{  
     if( csqa1[i] > 0 && csqa1[i] != '\0') { 
       fprintf( stream, "%c", csqa1[i] ); 
       k++;
	 }
     i++;
   } while ( csqa1[i] != '\0' && (k<imin) );

   fprintf( stream, " \n target   > " );

   i=0; k=0;
   do{ 
     if( csqa2[i] > 0 && csqa2[i] != '\0')  { 
 	   fprintf( stream, "%c", csqa2[i] ); 
       k++;
	 }
     i++;
   } while ( csqa2[i] != '\0' && (k<imin));

   fprintf( stream, "\n" );

   fprintf( stream, "\nIXYZ x_oldk, str. 1 original (PDB standard format)\n");
   for(i=0;i<rows;i++) {
	    fprintf( stream, "%4s%7d%4s  %c%c%c %1s%4d%12.3f %8.3f %8.3f\n", "ATOM",i+1,"CA",
	    	  aa_que[i][0], aa_que[i][1], aa_que[i][2],"L",i+1,
		      x_que_original[1][i],  x_que_original[2][i], x_que_original[3][i]);
   }
   fprintf( stream, "IXYZ DONE\n\n");
   fprintf( stream, "IXYZ x_old, str. 1 rotated (PDB standard format)\n");
   for(i=0;i<rows;i++) {
      fprintf( stream, "%4s%7d%4s  %c%c%c %1s%4d%12.3f %8.3f %8.3f\n", "ATOM",i+1,"CA",
         	  aa_que[i][0], aa_que[i][1], aa_que[i][2],"L",i+1,
		      x_que_sav[1][i], x_que_sav[2][i], x_que_sav[3][i]);
   }
   fprintf( stream, "IXYZ DONE\n\n");
   fprintf( stream, "IXYZ x_new, str. 2 original (PDB standard format)\n");
   for(i=0;i<cols;i++) {
      fprintf( stream, "%4s%7d%4s  %c%c%c %1s%4d%12.3f %8.3f %8.3f\n", "ATOM",i+1,"CA",
	    	  aa_tar[i][0], aa_tar[i][1], aa_tar[i][2],"L",i+1,
		      x_tar[1][i],  x_tar[2][i], x_tar[3][i]);
   }
   fprintf( stream, "IXYZ DONE\n\n");
   /*
   EQUIV i seqai : k : j seqaj ja
   EQUIV   105 GLN  105  :    86 :     98 VAL   98
   */  
   fprintf( stream, "EQUIV i seqa i : k : j seqa j \n");	
   for(i=0;i<nEqPairs_sav;i++) {	 
      fprintf( stream, "EQUIV %5d %c%c%c %4d  : %5d :  %5d  %c%c%c%4d\n",
		iina_sav[i],
		aa_que[iina_sav[i]-1][0], aa_que[iina_sav[i]-1][1], aa_que[iina_sav[i]-1][2],
		iina_sav[i], i,
		jinb_sav[i],
	    aa_tar[jinb_sav[i]-1][0], aa_tar[jinb_sav[i]-1][1], aa_tar[jinb_sav[i]-1][2],
		jinb_sav[i]);
   }
   fprintf( stream, "EQUIV DONE \n");	
   fclose( stream );
   } 

   free_cVector(csqa1,0,2*MAXSEQ-1);
   free_cVector(csqa2,0,2*MAXSEQ-1);
   free_cVector(csqam,0,2*MAXSEQ-1);
 } /* if(structal) */

/* FREE ALL ALLOCATED MEMORY: */
 free_iVector(iina_sav,0, MAXSEQ);
 free_iVector(jinb_sav,0, MAXSEQ);

 free_dMatrix(s,0,rows-1,0,cols-1);
 free_dMatrix(SS,0,rows-1,0,cols-1);
 free_dMatrix(TrRow,0,rows-1,0,cols-1);
 free_dMatrix(TrCol,0,rows-1,0,cols-1);
 free_iVector(jOpnPen,0, cols-1);
 free_iVector(jExtPen,0, cols-1);
 free_iVector(iOpnPen,0, rows-1);
 free_iVector(iExtPen,0, rows-1);

 free_dMatrix(x_que,1,3,0,rows-1);
 free_dMatrix(x_que_sav,1,3,0,rows-1);
 free_dMatrix(x_que_original,1,3,0,rows-1);
 free_dMatrix(x_tar,1,3,0,cols-1);

 free_cVector(a_que,0,rows-1);
 free_cVector(a_tar,0,cols-1);

 free_cVector(s1,1,StringMax);
 free_cVector(s2,1,StringMax);
 free_cVector(s3,1,StringMax);
 free_cVector(s4,1,StringMax);

 free_cMatrix(aa_que,0,rows-1,0,2);
 free_cMatrix(aa_tar,0,cols-1,0,2);

 /* freeing memory from PMI alignment: */
 free_dVector(Tcm,1,3);
 free_dVector(Tzero,1,3);
 free_dVector(TMPvec,1,3);
 free_dVector(Tcm_que,1,3);
 free_dVector(Tcm_tar,1,3);
 free_dMatrix(Rtemp,1,3,1,3);
 free_dMatrix(Runity,1,3,1,3);

/* printf("RMS = %f for %d equivalence pairs\n",RMSmin,nEqPairs); */

 /* START 3D HMM CALCULATIONS BASED ON JUST WRITTEN que.dat FILE */
if( !(structal) ) {  /* OLD STRUCTAL CALL IS NOT COUPLED WITH 3DHMM CALCS */  
 getrms_val = getrms(&ret_hscore, RMSmin);
 printf("RMSmin, 3dhmmscore = %f, %f ",RMSmin, ret_hscore);
}
 return 0;
}
